# coding: utf-8

import sys
from typing import Any

import pytest  # type: ignore # NOQA
from roundtrip import save_and_run  # type: ignore # NOQA


def test_monster(tmpdir: Any) -> None:
    program_src = '''\
    import ruyaml
    from textwrap import dedent

    class Monster(ruyaml.YAMLObject):
        yaml_tag = '!Monster'

        def __init__(self, name, hp, ac, attacks):
            self.name = name
            self.hp = hp
            self.ac = ac
            self.attacks = attacks

        def __repr__(self):
            return "%s(name=%r, hp=%r, ac=%r, attacks=%r)" % (
                self.__class__.__name__, self.name, self.hp, self.ac, self.attacks)

    yaml = ruyaml.YAML(typ='safe', pure='True')
    yaml = ruyaml.YAML()
    data = yaml.load(dedent("""\\
        --- !Monster
        name: Cave spider
        hp: [2,6]    # 2d6
        ac: 16
        attacks: [BITE, HURT]
    """))
    # normal dump, keys will be sorted
    from io import BytesIO
    buf = BytesIO()
    yaml.dump(data, buf)
    print(buf.getvalue().decode('utf-8'))
    assert buf.getvalue().decode('utf8') == dedent("""\\
        !Monster
        name: Cave spider
        hp: [2, 6]   # 2d6
        ac: 16
        attacks: [BITE, HURT]
    """)
    '''
    assert save_and_run(program_src, tmpdir) == 0


@pytest.mark.skipif(sys.version_info < (3, 0), reason='no __qualname__')  # type: ignore
def test_qualified_name00(tmpdir: Any) -> None:
    """issue 214"""
    program_src = """\
    import ruyaml
    from io import StringIO

    class A:
        def f(self):
            pass

    yaml = ruyaml.YAML(typ='unsafe', pure=True)
    yaml.explicit_end = True
    buf = StringIO()
    yaml.dump(A.f, buf)
    res = buf.getvalue()
    print('res', repr(res))
    assert res == "!!python/name:__main__.A.f ''\\n...\\n"
    x = ruyaml.load(res)
    assert x == A.f
    """
    assert save_and_run(program_src, tmpdir) == 0


@pytest.mark.skipif(sys.version_info < (3, 0), reason='no __qualname__')  # type: ignore
def test_qualified_name01(tmpdir: Any) -> None:
    """issue 214"""
    from io import StringIO

    import ruyaml.comments
    from ruyaml import YAML

    with pytest.warns(PendingDeprecationWarning):
        yaml = YAML(typ='unsafe', pure=True)
    yaml.explicit_end = True
    buf = StringIO()
    yaml.dump(ruyaml.comments.CommentedBase.yaml_anchor, buf)
    res = buf.getvalue()
    assert res == "!!python/name:ruyaml.comments.CommentedBase.yaml_anchor ''\n...\n"
    x = yaml.load(res)
    assert x == ruyaml.comments.CommentedBase.yaml_anchor



# coding: utf-8

from typing import Any

import pytest  # type: ignore # NOQA
from roundtrip import YAML, round_trip, round_trip_load  # type: ignore


def register_xxx(**kw: Any) -> None:
    import ruyaml as yaml

    class XXX(yaml.comments.CommentedMap):
        @staticmethod
        def yaml_dump(dumper: Any, data: Any) -> Any:
            return dumper.represent_mapping('!xxx', data)

        @classmethod
        def yaml_load(cls, constructor: Any, node: Any) -> Any:
            data = cls()
            yield data
            constructor.construct_mapping(node, data)

    yaml.add_constructor('!xxx', XXX.yaml_load, constructor=yaml.RoundTripConstructor)
    yaml.add_representer(XXX, XXX.yaml_dump, representer=yaml.RoundTripRepresenter)


class TestIndentFailures:
    def test_tag(self) -> None:
        round_trip(
            """\
        !!python/object:__main__.Developer
        name: Anthon
        location: Germany
        language: python
        """,
        )

    def test_full_tag(self) -> None:
        round_trip(
            """\
        !!tag:yaml.org,2002:python/object:__main__.Developer
        name: Anthon
        location: Germany
        language: python
        """,
        )

    def test_standard_tag(self) -> None:
        round_trip(
            """\
        !!tag:yaml.org,2002:python/object:map
        name: Anthon
        location: Germany
        language: python
        """,
        )

    def test_Y1(self) -> None:
        round_trip(
            """\
        !yyy
        name: Anthon
        location: Germany
        language: python
        """,
        )

    def test_Y2(self) -> None:
        round_trip(
            """\
        !!yyy
        name: Anthon
        location: Germany
        language: python
        """,
        )

    # @pytest.mark.xfail(strict=True)  # type: ignore
    def test_spec_6_26_tag_shorthands(self) -> None:
        from io import StringIO
        from textwrap import dedent

        from ruyaml import YAML

        inp = dedent(
            """\
        %TAG !e! tag:example.com,2000:app/
        ---
        - !local foo
        - !!str bar
        - !e!tag%21 baz
        """,
        )
        yaml = YAML()
        data = yaml.load(inp)
        buf = StringIO()
        yaml.dump(data, buf)
        print('buf:\n', buf.getvalue(), sep='')
        assert buf.getvalue() == inp


class TestTagGeneral:
    def test_unknow_handle(self) -> None:
        from ruyaml.parser import ParserError

        with pytest.raises(ParserError):
            round_trip(
                """\
            %TAG !x! tag:example.com,2000:app/
            ---
            - !y!tag%21 baz
            """,
            )


class TestRoundTripCustom:
    def test_X1(self) -> None:
        register_xxx()
        round_trip(
            """\
        !xxx
        name: Anthon
        location: Germany
        language: python
        """,
        )

    @pytest.mark.xfail(strict=True)  # type: ignore
    def test_X_pre_tag_comment(self) -> None:
        register_xxx()
        round_trip(
            """\
        -
          # hello
          !xxx
          name: Anthon
          location: Germany
          language: python
        """,
        )

    @pytest.mark.xfail(strict=True)  # type: ignore
    def test_X_post_tag_comment(self) -> None:
        register_xxx()
        round_trip(
            """\
        - !xxx
          # hello
          name: Anthon
          location: Germany
          language: python
        """,
        )

    def test_scalar_00(self) -> None:
        # https://stackoverflow.com/a/45967047/1307905
        round_trip(
            """\
        Outputs:
          Vpc:
            Value: !Ref: vpc    # first tag
            Export:
              Name: !Sub "${AWS::StackName}-Vpc"  # second tag
        """,
        )


class TestIssue201:
    def test_encoded_unicode_tag(self) -> None:
        round_trip_load(
            """
        s: !!python/%75nicode 'abc'
        """,
        )


class TestImplicitTaggedNodes:
    def test_scalar(self) -> None:
        data = round_trip(
            """\
        - !SString abcdefg
        - !SFloat 1.0
        - !SInt 1961
        - !SBool true
        - !SLit |
          glitter in the dark near the Tanhäuser gate
        """,
        )
        # tagged scalers have string or string types as value
        assert data[0].count('d') == 1
        assert data[1].count('1') == 1
        assert data[2].count('1') == 2
        assert data[3].count('u') == 1
        assert data[4].count('a') == 4

    def test_mapping(self) -> None:
        round_trip(
            """\
        - !Mapping {a: 1, b: 2}
        """,
        )

    def test_sequence(self) -> None:
        yaml = YAML()
        yaml.brace_single_entry_mapping_in_flow_sequence = True
        yaml.mapping_value_align = True
        yaml.round_trip(
            """
        - !Sequence [a, {b: 1}, {c: {d: 3}}]
        """,
        )

    def test_sequence2(self) -> None:
        yaml = YAML()
        yaml.mapping_value_align = True
        yaml.round_trip(
            """
        - !Sequence [a, b: 1, c: {d: 3}]
        """,
        )



# coding: utf-8

"""
collections.OrderedDict is a new class not supported by PyYAML (issue 83 by Frazer McLean)

This is now so integrated in Python that it can be mapped to !!omap

"""

import pytest  # type: ignore  # NOQA
from roundtrip import (  # type: ignore # NOQA
    dedent,
    round_trip,
    round_trip_dump,
    round_trip_load,
)


class TestOrderedDict:
    def test_ordereddict(self) -> None:
        from collections import OrderedDict

        assert round_trip_dump(OrderedDict()) == '!!omap []\n'



import pytest  # type: ignore  # NOQA

from ruyaml.docinfo import DocInfo, Tag, Version, version  # NOQA


class TestVersion:
    def test_create_from_integers(self) -> None:
        v = Version(1, 2)
        assert v.major == 1
        assert v.minor == 2

    def test_create_using_generator(self) -> None:
        v = version(1, 2)
        assert isinstance(v, Version)
        assert v.major == 1
        assert v.minor == 2

    def test_create_from_string_using_generator(self) -> None:
        v = version('1.2')
        assert isinstance(v, Version)
        assert v.major == 1
        assert v.minor == 2

    def test_create_from_string_extra_param(self) -> None:
        with pytest.raises(AssertionError):
            _ = version('1.2', 3)

    def test_create_from_single_integer(self) -> None:
        with pytest.raises(AssertionError):
            _ = version(1)
        with pytest.raises(TypeError):
            _ = Version(1)  # type: ignore


class TestDocInfo:
    def test_empty(self) -> None:
        di = DocInfo()
        assert di.requested_version is None
        assert di.doc_version is None
        assert di.tags == []

    def test_versions(self) -> None:
        di = DocInfo(version('1.2'), version('1.1'))
        assert di.requested_version > di.doc_version  # type: ignore



# coding: utf-8

import pytest  # type: ignore  # NOQA
from roundtrip import (  # type: ignore
    round_trip,
    round_trip_dump_all,
    round_trip_load_all,
)


class TestDocument:
    def test_single_doc_begin_end(self) -> None:
        inp = """\
        ---
        - a
        - b
        ...
        """
        round_trip(inp, explicit_start=True, explicit_end=True)

    def test_multi_doc_begin_end(self) -> None:
        inp = """\
        ---
        - a
        ...
        ---
        - b
        ...
        """
        docs = list(round_trip_load_all(inp))
        assert docs == [['a'], ['b']]
        out = round_trip_dump_all(docs, explicit_start=True, explicit_end=True)
        assert out == '---\n- a\n...\n---\n- b\n...\n'

    def test_multi_doc_no_start(self) -> None:
        inp = """\
        - a
        ...
        ---
        - b
        ...
        """
        docs = list(round_trip_load_all(inp))
        assert docs == [['a'], ['b']]

    def test_multi_doc_no_end(self) -> None:
        inp = """\
        - a
        ---
        - b
        """
        docs = list(round_trip_load_all(inp))
        assert docs == [['a'], ['b']]

    def test_multi_doc_ends_only(self) -> None:
        # this is ok in 1.2
        inp = """\
        - a
        ...
        - b
        ...
        """
        docs = list(round_trip_load_all(inp, version=(1, 2)))
        assert docs == [['a'], ['b']]

    def test_single_scalar_comment(self) -> None:
        import ruyaml as yaml

        inp = """\
        one # comment
        two
        """
        with pytest.raises(yaml.parser.ParserError):
            d = list(round_trip_load_all(inp, version=(1, 2)))  # NOQA

    def test_scalar_after_seq_document(self) -> None:
        import ruyaml as yaml

        inp = """\
        [ 42 ]
        hello
        """
        with pytest.raises(yaml.parser.ParserError):
            d = list(round_trip_load_all(inp, version=(1, 2)))  # NOQA

    def test_yunk_after_explicit_document_end(self) -> None:
        import ruyaml as yaml

        inp = """\
        hello: world
        ... this is no comment
        """
        with pytest.raises(yaml.parser.ParserError):
            d = list(round_trip_load_all(inp, version=(1, 2)))  # NOQA

    def test_multi_doc_ends_only_1_1(self) -> None:
        import ruyaml as yaml

        # this is not ok in 1.1
        with pytest.raises(yaml.parser.ParserError):
            inp = """\
            - a
            ...
            - b
            ...
            """
            docs = list(round_trip_load_all(inp, version=(1, 1)))
            assert docs == [['a'], ['b']]  # not True, but not reached



# Skipped because we have no idea where this "canonical" module
# comes from, nor where all those fixtures originate
import pytest

import ruyaml

# import canonical  # NOQA


def test_canonical_scanner(canonical_filename, verbose=False):
    with open(canonical_filename, 'rb') as fp0:
        data = fp0.read()
    tokens = list(ruyaml.canonical_scan(data))
    assert tokens, tokens
    if verbose:
        for token in tokens:
            print(token)


test_canonical_scanner.unittest = ['.canonical']


def test_canonical_parser(canonical_filename, verbose=False):
    with open(canonical_filename, 'rb') as fp0:
        data = fp0.read()
    events = list(ruyaml.canonical_parse(data))
    assert events, events
    if verbose:
        for event in events:
            print(event)


test_canonical_parser.unittest = ['.canonical']


def test_canonical_error(data_filename, canonical_filename, verbose=False):
    with open(data_filename, 'rb') as fp0:
        data = fp0.read()
    try:
        output = list(ruyaml.canonical_load_all(data))  # NOQA
    except ruyaml.YAMLError as exc:
        if verbose:
            print(exc)
    else:
        raise AssertionError('expected an exception')


test_canonical_error.unittest = ['.data', '.canonical']
test_canonical_error.skip = ['.empty']

if __name__ == '__main__':
    import test_appliance

    test_appliance.run(globals())



# coding: utf-8

from typing import Any

import pytest  # type: ignore  # NOQA
from roundtrip import (  # type: ignore  # NOQA
    YAML,
    dedent,
    round_trip,
    round_trip_dump,
    round_trip_load,
)


def rt(s: str) -> str:
    res = round_trip_dump(round_trip_load(s))
    assert res is not None
    return res.strip() + '\n'  # type: ignore


class TestIndent:
    def test_roundtrip_inline_list(self) -> None:
        s = 'a: [a, b, c]\n'
        output = rt(s)
        assert s == output

    def test_roundtrip_mapping_of_inline_lists(self) -> None:
        s = dedent(
            """\
        a: [a, b, c]
        j: [k, l, m]
        """
        )
        output = rt(s)
        assert s == output

    def test_roundtrip_mapping_of_inline_lists_comments(self) -> None:
        s = dedent(
            """\
        # comment A
        a: [a, b, c]
        # comment B
        j: [k, l, m]
        """
        )
        output = rt(s)
        assert s == output

    def test_roundtrip_mapping_of_inline_sequence_eol_comments(self) -> None:
        s = dedent(
            """\
        # comment A
        a: [a, b, c]  # comment B
        j: [k, l, m]  # comment C
        """
        )
        output = rt(s)
        assert s == output

    # first test by explicitly setting flow style
    def test_added_inline_list(self) -> None:
        s1 = dedent(
            """
        a:
        - b
        - c
        - d
        """
        )
        s = 'a: [b, c, d]\n'
        data = round_trip_load(s1)
        val = data['a']
        val.fa.set_flow_style()
        # print(type(val), '_yaml_format' in dir(val))
        output = round_trip_dump(data)
        assert s == output

    # ############ flow mappings

    def test_roundtrip_flow_mapping(self) -> None:
        s = dedent(
            """\
        - {a: 1, b: hallo}
        - {j: fka, k: 42}
        """
        )
        data = round_trip_load(s)
        output = round_trip_dump(data)
        assert s == output

    def test_roundtrip_sequence_of_inline_mappings_eol_comments(self) -> None:
        s = dedent(
            """\
        # comment A
        - {a: 1, b: hallo}  # comment B
        - {j: fka, k: 42}  # comment C
        """
        )
        output = rt(s)
        assert s == output

    def test_indent_top_level(self) -> None:
        inp = """
        -   a:
            -   b
        """
        round_trip(inp, indent=4)

    def test_set_indent_5_block_list_indent_1(self) -> None:
        inp = """
        a:
         -   b: c
         -   1
         -   d:
              -   2
        """
        round_trip(inp, indent=5, block_seq_indent=1)

    def test_set_indent_4_block_list_indent_2(self) -> None:
        inp = """
        a:
          - b: c
          - 1
          - d:
              - 2
        """
        round_trip(inp, indent=4, block_seq_indent=2)

    def test_set_indent_3_block_list_indent_0(self) -> None:
        inp = """
        a:
        -  b: c
        -  1
        -  d:
           -  2
        """
        round_trip(inp, indent=3, block_seq_indent=0)

    def Xtest_set_indent_3_block_list_indent_2(self) -> None:
        inp = """
        a:
          -
           b: c
          -
           1
          -
           d:
             -
              2
        """
        round_trip(inp, indent=3, block_seq_indent=2)

    def test_set_indent_3_block_list_indent_2(self) -> None:
        inp = """
        a:
          - b: c
          - 1
          - d:
             - 2
        """
        round_trip(inp, indent=3, block_seq_indent=2)

    def Xtest_set_indent_2_block_list_indent_2(self) -> None:
        inp = """
        a:
          -
           b: c
          -
           1
          -
           d:
             -
              2
        """
        round_trip(inp, indent=2, block_seq_indent=2)

    # this is how it should be: block_seq_indent stretches the indent
    def test_set_indent_2_block_list_indent_2(self) -> None:
        inp = """
        a:
          - b: c
          - 1
          - d:
            - 2
        """
        round_trip(inp, indent=2, block_seq_indent=2)

    # have to set indent!
    def test_roundtrip_four_space_indents(self) -> None:
        # fmt: off
        s = (
            'a:\n'
            '-   foo\n'
            '-   bar\n'
        )
        # fmt: on
        round_trip(s, indent=4)

    def test_roundtrip_four_space_indents_no_fail(self) -> None:
        inp = """
        a:
        -   foo
        -   bar
        """
        exp = """
        a:
        - foo
        - bar
        """
        assert round_trip_dump(round_trip_load(inp)) == dedent(exp)


class TestYpkgIndent:
    def test_00(self) -> None:
        inp = """
        name       : nano
        version    : 2.3.2
        release    : 1
        homepage   : http://www.nano-editor.org
        source     :
          - http://www.nano-editor.org/dist/v2.3/nano-2.3.2.tar.gz :<SPC>
                ff30924807ea289f5b60106be8
        license    : GPL-2.0
        summary    : GNU nano is an easy-to-use text editor
        builddeps  :
          - ncurses-devel
        description: |
            GNU nano is an easy-to-use text editor originally designed
            as a replacement for Pico, the ncurses-based editor from the non-free mailer
            package Pine (itself now available under the Apache License as Alpine).
        """
        round_trip(
            inp,
            indent=4,
            block_seq_indent=2,
            top_level_colon_align=True,
            prefix_colon=' ',
        )


def guess(s: str) -> Any:
    from ruyaml.util import load_yaml_guess_indent

    x, y, z = load_yaml_guess_indent(dedent(s))
    return y, z


class TestGuessIndent:
    def test_guess_20(self) -> None:
        inp = """\
        a:
        - 1
        """
        assert guess(inp) == (2, 0)

    def test_guess_42(self) -> None:
        inp = """\
        a:
          - 1
        """
        assert guess(inp) == (4, 2)

    def test_guess_42a(self) -> None:
        # block seq indent prevails over nested key indent level
        inp = """\
        b:
              a:
                - 1
        """
        assert guess(inp) == (4, 2)

    def test_guess_3None(self) -> None:
        inp = """\
        b:
           a: 1
        """
        assert guess(inp) == (3, None)

    def test_guess_with_preserve_quotes(self) -> None:
        from ruyaml.scalarstring import DoubleQuotedScalarString
        from ruyaml.util import load_yaml_guess_indent

        inp = """\
        b:
           a: "hello world"
        """
        yaml = YAML()
        yaml.preserve_quotes = True
        x, y, z = load_yaml_guess_indent(dedent(inp), yaml=yaml)
        assert y == 3
        assert z is None
        assert isinstance(x['b']['a'], DoubleQuotedScalarString)


class TestSeparateMapSeqIndents:
    # using uncommon 6 indent with 3 push in as 2 push in automatically
    # gets you 4 indent even if not set
    def test_00(self) -> None:
        # old style
        yaml = YAML()
        yaml.indent = 6
        yaml.block_seq_indent = 3
        inp = """
        a:
           -  1
           -  [1, 2]
        """
        yaml.round_trip(inp)

    def test_01(self) -> None:
        yaml = YAML()
        yaml.indent(sequence=6)
        yaml.indent(offset=3)
        inp = """
        a:
           -  1
           -  {b: 3}
        """
        yaml.round_trip(inp)

    def test_02(self) -> None:
        yaml = YAML()
        yaml.indent(mapping=5, sequence=6, offset=3)
        inp = """
        a:
             b:
                -  1
                -  [1, 2]
        """
        yaml.round_trip(inp)

    def test_03(self) -> None:
        inp = """
        a:
            b:
                c:
                -   1
                -   [1, 2]
        """
        round_trip(inp, indent=4)

    def test_04(self) -> None:
        yaml = YAML()
        yaml.indent(mapping=5, sequence=6)
        inp = """
        a:
             b:
             -     1
             -     [1, 2]
             -     {d: 3.14}
        """
        yaml.round_trip(inp)

    def test_issue_51(self) -> None:
        yaml = YAML()
        # yaml.map_indent = 2 # the default
        yaml.indent(sequence=4, offset=2)
        yaml.preserve_quotes = True
        yaml.round_trip(
            """
        role::startup::author::rsyslog_inputs:
          imfile:
            - ruleset: 'AEM-slinglog'
              File: '/opt/aem/author/crx-quickstart/logs/error.log'
              startmsg.regex: '^[-+T.:[:digit:]]*'
              tag: 'error'
            - ruleset: 'AEM-slinglog'
              File: '/opt/aem/author/crx-quickstart/logs/stdout.log'
              startmsg.regex: '^[-+T.:[:digit:]]*'
              tag: 'stdout'
        """
        )


# ############ indentation



# coding: utf-8

import platform
import sys
from textwrap import dedent

import pytest  # type: ignore  # NOQA

NO_CLIB_VER = (3, 12)


@pytest.mark.skipif(  # type: ignore
    platform.python_implementation() in ['Jython', 'PyPy'],
    reason='Jython throws RepresenterError',
)
def test_load_cyaml() -> None:
    print("???????????????????????", platform.python_implementation())
    import ruyaml

    if sys.version_info >= NO_CLIB_VER:
        return
    yaml = ruyaml.YAML(typ='safe', pure=False)
    assert ruyaml.__with_libyaml__

    yaml.load('abc: 1')


@pytest.mark.skipif(
    sys.version_info >= NO_CLIB_VER  # type: ignore
    or platform.python_implementation() in ['Jython', 'PyPy'],
    reason='no _PyGC_FINALIZED',
)
def test_dump_cyaml() -> None:
    import ruyaml

    if sys.version_info >= NO_CLIB_VER:
        return
    data = {'a': 1, 'b': 2}
    yaml = ruyaml.YAML(typ='safe', pure=False)
    yaml.default_flow_style = False
    yaml.allow_unicode = True
    buf = ruyaml.compat.StringIO()
    yaml.dump(data, buf)
    assert buf.getvalue() == 'a: 1\nb: 2\n'


@pytest.mark.skipif(  # type: ignore
    platform.python_implementation() in ['Jython', 'PyPy'],
    reason='not avialable',
)
def test_load_cyaml_1_2() -> None:
    # issue 155
    import ruyaml

    if sys.version_info >= NO_CLIB_VER:
        return
    assert ruyaml.__with_libyaml__
    inp = dedent(
        """\
    %YAML 1.2
    ---
    num_epochs: 70000
    """
    )
    yaml = ruyaml.YAML(typ='safe')
    yaml.load(inp)


@pytest.mark.skipif(  # type: ignore
    platform.python_implementation() in ['Jython', 'PyPy'],
    reason='not available',
)
def test_dump_cyaml_1_2() -> None:
    # issue 155
    from io import StringIO

    import ruyaml

    if sys.version_info >= NO_CLIB_VER:
        return
    assert ruyaml.__with_libyaml__
    yaml = ruyaml.YAML(typ='safe')
    yaml.version = (1, 2)
    yaml.default_flow_style = False
    data = {'a': 1, 'b': 2}
    exp = dedent(
        """\
    %YAML 1.2
    ---
    a: 1
    b: 2
    """
    )
    buf = StringIO()
    yaml.dump(data, buf)
    assert buf.getvalue() == exp



import pytest


def _fn(r, n):
    raise pytest.skip.Exception("XXX_filename fixtures are unknown")


@pytest.fixture(scope="function")
def canonical_filename(request):
    return _fn(request, "canonical")


@pytest.fixture(scope="function")
def data_filename(request):
    return _fn(request, "data")


@pytest.fixture(scope="function")
def detect_filename(request):
    return _fn(request, "detect")


@pytest.fixture(scope="function")
def events_filename(request):
    return _fn(request, "events")



# coding: utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

if False:  # MYPY
    from typing import Any, Dict  # NOQA

_package_data = dict(
    full_package_name='ruyaml',
    version_info=(0, 18, 15),
    __version__='0.18.15',
    author='ruyaml contributors',
    author_email='none.yet@github.org',
    description='ruyaml is a YAML parser/emitter that supports roundtrip preservation of comments, seq/map flow style, and map key order',  # NOQA
    entry_points=None,
    since=2014,
    extras_require={
        ':platform_python_implementation=="CPython" and python_version<"3.8"': [
            'ruyaml.clib>=0.1.2',
        ],
    },
)  # type: Dict[Any, Any]


version_info = _package_data['version_info']
__version__ = _package_data['__version__']

try:
    from .cyaml import *  # NOQA

    __with_libyaml__ = True
except (ImportError, ValueError):  # for Jython
    __with_libyaml__ = False

from ruyaml.main import *  # NOQA



# coding: utf-8

import sys

import pytest  # type:ignore  # NOQA

last_to_warn = (0, 17, 40)


@pytest.mark.skipif(
    sys.version_info < (3, 7) or sys.version_info >= (3, 9),  # type: ignore
    reason='collections not available?',
)
def test_collections_deprecation() -> None:
    with pytest.warns(DeprecationWarning):
        from collections import Hashable  # type: ignore  # NOQA


class TestFunctionDeprecation:
    def test_deprecation_scan(self) -> None:
        import ruyaml

        if ruyaml.version_info <= last_to_warn:
            with pytest.warns(PendingDeprecationWarning):
                data = ruyaml.load('a: 42')  # NOQA
        else:
            with pytest.raises(AttributeError):
                data = ruyaml.load('a: 42')  # NOQA

    def test_deprecation_parse(self) -> None:
        import ruyaml

        if ruyaml.version_info <= last_to_warn:
            data = ruyaml.parse('a: 42')  # NOQA
        else:
            with pytest.raises(AttributeError):
                data = ruyaml.parse('a: 42')  # NOQA

    def test_deprecation_compose(self) -> None:
        import ruyaml

        if ruyaml.version_info <= last_to_warn:
            with pytest.warns(PendingDeprecationWarning):
                data = ruyaml.compose('a: 42')  # NOQA
        else:
            with pytest.raises(AttributeError):
                data = ruyaml.parse('a: 42')  # NOQA

    def test_deprecation_compose_all(self) -> None:
        import ruyaml

        if ruyaml.version_info <= last_to_warn:
            data = ruyaml.compose_all('a: 42')  # NOQA
        else:
            with pytest.raises(AttributeError):
                data = ruyaml.parse('a: 42')  # NOQA

    def test_deprecation_load(self) -> None:
        import ruyaml

        if ruyaml.version_info <= last_to_warn:
            with pytest.warns(PendingDeprecationWarning):
                data = ruyaml.load('a: 42')  # NOQA
        else:
            with pytest.raises(AttributeError):
                data = ruyaml.parse('a: 42')  # NOQA

    def test_deprecation_load_all(self) -> None:
        import ruyaml

        if ruyaml.version_info <= last_to_warn:
            data = ruyaml.load_all('a: 42')  # NOQA
        else:
            with pytest.raises(AttributeError):
                data = ruyaml.parse('a: 42')  # NOQA

    def test_deprecation_safe_load(self) -> None:
        import ruyaml

        if ruyaml.version_info <= last_to_warn:
            with pytest.warns(PendingDeprecationWarning):
                data = ruyaml.safe_load('a: 42')  # NOQA
        else:
            with pytest.raises(AttributeError):
                data = ruyaml.parse('a: 42')  # NOQA

    def test_deprecation_round_trip_load(self) -> None:
        import ruyaml

        if ruyaml.version_info <= last_to_warn:
            with pytest.warns(PendingDeprecationWarning):
                data = ruyaml.round_trip_load('a: 42')  # NOQA
        else:
            with pytest.raises(AttributeError):
                data = ruyaml.parse('a: 42')  # NOQA


class TestYamlTyp:
    def test_unsafe_deprecation(self) -> None:
        import ruyaml

        if ruyaml.version_info < (0, 18, 0):
            yaml = ruyaml.YAML(typ='unsafe')
        else:
            with pytest.warns(PendingDeprecationWarning):
                #   with pytest.raises(SystemExit):
                yaml = ruyaml.YAML(typ='unsafe')  # NOQA

    def test_full_load_error(self) -> None:
        import ruyaml

        yaml = ruyaml.YAML(typ='full', pure=True)
        with pytest.raises(ruyaml.error.YAMLError):
            yaml.load('a: b')
        yaml = ruyaml.YAML(typ='full')  # C scanner/loader
        with pytest.raises(ruyaml.error.YAMLError):
            yaml.load('a: b')

    def test_full_rt(self) -> None:
        import io
        import os

        import ruyaml

        yaml = ruyaml.YAML(typ='full', pure=True)
        buf = io.BytesIO()
        yaml.dump([{'fun': os.system}], buf)
        print(buf.getvalue())
        yaml = ruyaml.YAML()
        data = yaml.load(buf.getvalue())
        print(data)
        ts = data[0]['fun']
        assert 'posix.system' in str(ts.tag)



# coding: utf-8

"""
helper routines for testing round trip of commented YAML data
"""
import io
import sys
import textwrap
from pathlib import Path
from typing import Any, Optional, Union

import ruyaml

unset = object()


def dedent(data: str) -> str:
    try:
        position_of_first_newline = data.index('\n')
        for idx in range(position_of_first_newline):
            if not data[idx].isspace():
                raise ValueError
    except ValueError:
        pass
    else:
        data = data[position_of_first_newline + 1 :]
    return textwrap.dedent(data)


def round_trip_load(
    inp: Any,
    preserve_quotes: Optional[bool] = None,
    version: Optional[Any] = None,
) -> Any:
    import ruyaml  # NOQA

    dinp = dedent(inp).replace('<SPC>', ' ')  # to allow for end-of-line spaces
    yaml = ruyaml.YAML()
    yaml.preserve_quotes = preserve_quotes
    yaml.version = version
    return yaml.load(dinp)


def round_trip_load_all(
    inp: Any,
    preserve_quotes: Optional[bool] = None,
    version: Optional[Any] = None,
) -> Any:
    import ruyaml  # NOQA

    dinp = dedent(inp).replace('<SPC>', ' ')  # to allow for end-of-line spaces
    yaml = ruyaml.YAML()
    yaml.preserve_quotes = preserve_quotes
    yaml.version = version
    return yaml.load_all(dinp)


def round_trip_dump(
    data: Any,
    stream: Any = None,  # *,
    indent: Optional[int] = None,
    block_seq_indent: Optional[int] = None,
    default_flow_style: Any = unset,
    top_level_colon_align: Any = None,
    prefix_colon: Any = None,
    explicit_start: Optional[bool] = None,
    explicit_end: Optional[bool] = None,
    version: Optional[Any] = None,
    allow_unicode: bool = True,
) -> Union[str, None]:
    import ruyaml  # NOQA

    yaml = ruyaml.YAML()
    yaml.indent(mapping=indent, sequence=indent, offset=block_seq_indent)
    if default_flow_style is not unset:
        yaml.default_flow_style = default_flow_style
    yaml.top_level_colon_align = top_level_colon_align
    yaml.prefix_colon = prefix_colon
    yaml.explicit_start = explicit_start
    yaml.explicit_end = explicit_end
    yaml.version = version
    yaml.allow_unicode = allow_unicode
    if stream is not None:
        yaml.dump(data, stream=stream)
        return None
    buf = io.StringIO()
    yaml.dump(data, stream=buf)
    return buf.getvalue()


def round_trip_dump_all(
    data: Any,
    stream: Any = None,  # *,
    indent: Optional[int] = None,
    block_seq_indent: Optional[int] = None,
    default_flow_style: Any = unset,
    top_level_colon_align: Any = None,
    prefix_colon: Any = None,
    explicit_start: Optional[bool] = None,
    explicit_end: Optional[bool] = None,
    version: Optional[Any] = None,
    allow_unicode: bool = True,
) -> Union[str, None]:
    import ruyaml  # NOQA

    yaml = ruyaml.YAML()
    yaml.indent(mapping=indent, sequence=indent, offset=block_seq_indent)
    if default_flow_style is not unset:
        yaml.default_flow_style = default_flow_style
    yaml.top_level_colon_align = top_level_colon_align
    yaml.prefix_colon = prefix_colon
    yaml.explicit_start = explicit_start
    yaml.explicit_end = explicit_end
    yaml.version = version
    yaml.allow_unicode = allow_unicode
    if stream is not None:
        yaml.dump(data, stream=stream)
        return None
    buf = io.StringIO()
    yaml.dump_all(data, stream=buf)
    return buf.getvalue()


def diff(inp: str, outp: str, file_name: str = 'stdin') -> None:
    import difflib

    inl = inp.splitlines(True)  # True for keepends
    outl = outp.splitlines(True)
    diff = difflib.unified_diff(inl, outl, file_name, 'round trip YAML')
    for line in diff:
        sys.stdout.write(line)


def round_trip(
    inp: str,
    outp: Optional[str] = None,
    extra: Optional[str] = None,
    intermediate: Any = None,
    indent: Optional[int] = None,
    block_seq_indent: Optional[int] = None,
    default_flow_style: Any = unset,
    top_level_colon_align: Any = None,
    prefix_colon: Any = None,
    preserve_quotes: Any = None,
    explicit_start: Optional[bool] = None,
    explicit_end: Optional[bool] = None,
    version: Optional[Any] = None,
    dump_data: Any = None,
) -> Any:
    """
    inp:    input string to parse
    outp:   expected output (equals input if not specified)
    """
    if outp is None:
        outp = inp
    doutp = dedent(outp).replace('<SPC>', ' ')  # allow for end-of-line spaces
    if extra is not None:
        doutp += extra
    data = round_trip_load(inp, preserve_quotes=preserve_quotes)
    if dump_data:
        print('data', data)
    if intermediate is not None:
        if isinstance(intermediate, dict):
            for k, v in intermediate.items():
                if data[k] != v:
                    print('{0!r} <> {1!r}'.format(data[k], v))
                    raise ValueError
    res = round_trip_dump(
        data,
        indent=indent,
        block_seq_indent=block_seq_indent,
        top_level_colon_align=top_level_colon_align,
        prefix_colon=prefix_colon,
        explicit_start=explicit_start,
        explicit_end=explicit_end,
        version=version,
    )
    assert isinstance(res, str)
    if res != doutp:
        diff(doutp, res, 'input string')
    print('\nroundtrip data:\n', res, sep="")
    assert res == doutp
    res = round_trip_dump(
        data,
        indent=indent,
        block_seq_indent=block_seq_indent,
        top_level_colon_align=top_level_colon_align,
        prefix_colon=prefix_colon,
        explicit_start=explicit_start,
        explicit_end=explicit_end,
        version=version,
    )
    print('roundtrip second round data:\n', res, sep="")
    assert res == doutp
    return data


def na_round_trip(
    inp: str,
    outp: Optional[str] = None,
    extra: Optional[str] = None,
    intermediate: Any = None,
    indent: Optional[int] = None,
    top_level_colon_align: Any = None,
    prefix_colon: Any = None,
    preserve_quotes: Any = None,
    explicit_start: Optional[bool] = None,
    explicit_end: Optional[bool] = None,
    version: Optional[Any] = None,
    dump_data: Any = None,
) -> Any:
    """
    inp:    input string to parse
    outp:   expected output (equals input if not specified)
    """
    inp = dedent(inp)
    if outp is None:
        outp = inp
    if version is not None:
        version = version
    doutp = dedent(outp)
    if extra is not None:
        doutp += extra
    yaml = YAML()
    yaml.preserve_quotes = preserve_quotes
    yaml.scalar_after_indicator = False  # newline after every directives end
    data = yaml.load(inp)
    if dump_data:
        print('data', data)
    if intermediate is not None:
        if isinstance(intermediate, dict):
            for k, v in intermediate.items():
                if data[k] != v:
                    print('{0!r} <> {1!r}'.format(data[k], v))
                    raise ValueError
    yaml.indent = indent
    yaml.top_level_colon_align = top_level_colon_align
    yaml.prefix_colon = prefix_colon
    yaml.explicit_start = explicit_start
    yaml.explicit_end = explicit_end
    res = yaml.dump(data, compare=doutp)
    return res


def YAML(**kw: Any) -> Any:
    import ruyaml  # NOQA

    class MyYAML(ruyaml.YAML):
        """auto dedent string parameters on load"""

        def load(self, stream: Any) -> Any:
            if isinstance(stream, str):
                if stream and stream[0] == '\n':
                    stream = stream[1:]
                stream = textwrap.dedent(stream)
            return ruyaml.YAML.load(self, stream)

        def load_all(self, stream: Any) -> Any:
            if isinstance(stream, str):
                if stream and stream[0] == '\n':
                    stream = stream[1:]
                stream = textwrap.dedent(stream)
            for d in ruyaml.YAML.load_all(self, stream):
                yield d

        def dump(self, data: Any, **kw: Any) -> Any:  # type: ignore
            from ruyaml.compat import BytesIO, StringIO  # NOQA

            assert ('stream' in kw) ^ ('compare' in kw)
            if 'stream' in kw:
                return ruyaml.YAML.dump(data, **kw)
            lkw = kw.copy()
            expected = textwrap.dedent(lkw.pop('compare'))
            unordered_lines = lkw.pop('unordered_lines', False)
            if expected and expected[0] == '\n':
                expected = expected[1:]
            lkw['stream'] = st = StringIO()
            ruyaml.YAML.dump(self, data, **lkw)
            res = st.getvalue()
            print(res)
            if unordered_lines:
                res = sorted(res.splitlines())  # type: ignore
                expected = sorted(expected.splitlines())  # type: ignore
            assert res == expected

        def round_trip(self, stream: Any, **kw: Any) -> None:
            from ruyaml.compat import BytesIO, StringIO  # NOQA

            assert isinstance(stream, str)
            lkw = kw.copy()
            if stream and stream[0] == '\n':
                stream = stream[1:]
            stream = textwrap.dedent(stream)
            data = ruyaml.YAML.load(self, stream)
            outp = lkw.pop('outp', stream)
            lkw['stream'] = st = StringIO()
            ruyaml.YAML.dump(self, data, **lkw)
            res = st.getvalue()
            if res != outp:
                diff(outp, res, 'input string')
            assert res == outp

        def round_trip_all(self, stream: Any, **kw: Any) -> None:
            from ruyaml.compat import BytesIO, StringIO  # NOQA

            assert isinstance(stream, str)
            lkw = kw.copy()
            if stream and stream[0] == '\n':
                stream = stream[1:]
            stream = textwrap.dedent(stream)
            data = list(ruyaml.YAML.load_all(self, stream))
            outp = lkw.pop('outp', stream)
            lkw['stream'] = st = StringIO()
            ruyaml.YAML.dump_all(self, data, **lkw)
            res = st.getvalue()
            if res != outp:
                diff(outp, res, 'input string')
            assert res == outp

    return MyYAML(**kw)


def save_and_run(
    program: str,
    base_dir: Optional[Any] = None,
    output: Optional[Any] = None,
    file_name: Optional[Any] = None,
    optimized: bool = False,
) -> int:
    """
    safe and run a python program, thereby circumventing any restrictions on module level
    imports
    """
    from subprocess import STDOUT, CalledProcessError, check_output

    if not hasattr(base_dir, 'hash'):
        base_dir = Path(str(base_dir))
    if file_name is None:
        file_name = 'safe_and_run_tmp.py'
    file_name = base_dir / file_name  # type: ignore
    file_name.write_text(dedent(program))

    try:
        cmd = [sys.executable, '-Wd']
        if optimized:
            cmd.append('-O')
        cmd.append(str(file_name))
        print('running:', *cmd)
        # 3.5 needs strings
        res = check_output(
            cmd, stderr=STDOUT, universal_newlines=True, cwd=str(base_dir)
        )
        if output is not None:
            if '__pypy__' in sys.builtin_module_names:
                res1 = res.splitlines(True)
                res2 = [line for line in res1 if 'no version info' not in line]
                res = ''.join(res2)
            print('result:  ', res, end='')
            print('expected:', output, end='')
            assert res == output
    except CalledProcessError as exception:
        print("##### Running '{} {}' FAILED #####".format(sys.executable, file_name))
        print(exception.output)
        return exception.returncode
    return 0



# coding: utf-8

"""
http://yaml.org/type/timestamp.html specifies the regexp to use
for datetime.date and datetime.datetime construction. Date is simple
but datetime can have 'T' or 't' as well as 'Z' or a timezone offset (in
hours and minutes). This information was originally used to create
a UTC datetime and then discarded

examples from the above:

canonical:        2001-12-15T02:59:43.1Z
valid iso8601:    2001-12-14t21:59:43.10-05:00
space separated:  2001-12-14 21:59:43.10 -5
no time zone (Z): 2001-12-15 2:59:43.10
date (00:00:00Z): 2002-12-14

Please note that a fraction can only be included if not equal to 0

"""

import copy
import sys
from datetime import datetime as DateTime
from datetime import timedelta as TimeDelta
from datetime import timezone as TimeZone

import pytest  # type: ignore  # NOQA
from roundtrip import (  # type: ignore # NOQA
    dedent,
    round_trip,
    round_trip_dump,
    round_trip_load,
)


class TestDateTime:
    def test_date_only(self) -> None:
        inp = """
        - 2011-10-02
        """
        exp = """
        - 2011-10-02
        """
        round_trip(inp, exp)

    def test_zero_fraction(self) -> None:
        inp = """
        - 2011-10-02 16:45:00.0
        """
        exp = """
        - 2011-10-02 16:45:00
        """
        round_trip(inp, exp)

    def test_long_fraction(self) -> None:
        inp = """
        - 2011-10-02 16:45:00.1234      # expand with zeros
        - 2011-10-02 16:45:00.123456
        - 2011-10-02 16:45:00.12345612  # round to microseconds
        - 2011-10-02 16:45:00.1234565   # round up
        - 2011-10-02 16:45:00.12345678  # round up
        """
        exp = """
        - 2011-10-02 16:45:00.123400    # expand with zeros
        - 2011-10-02 16:45:00.123456
        - 2011-10-02 16:45:00.123456    # round to microseconds
        - 2011-10-02 16:45:00.123457    # round up
        - 2011-10-02 16:45:00.123457    # round up
        """
        round_trip(inp, exp)

    def test_canonical(self) -> None:
        inp = """
        - 2011-10-02T16:45:00.1Z
        """
        exp = """
        - 2011-10-02T16:45:00.100000Z
        """
        round_trip(inp, exp)

    def test_spaced_timezone(self) -> None:
        inp = """
        - 2011-10-02T11:45:00 -5
        """
        exp = """
        - 2011-10-02T11:45:00-5
        """
        round_trip(inp, exp)

    def test_normal_timezone(self) -> None:
        round_trip(
            """
        - 2011-10-02T11:45:00-5
        - 2011-10-02 11:45:00-5
        - 2011-10-02T11:45:00-05:00
        - 2011-10-02 11:45:00-05:00
        """
        )

    def test_no_timezone(self) -> None:
        inp = """
        - 2011-10-02 6:45:00
        """
        exp = """
        - 2011-10-02 06:45:00
        """
        round_trip(inp, exp)

    def test_explicit_T(self) -> None:
        inp = """
        - 2011-10-02T16:45:00
        """
        exp = """
        - 2011-10-02T16:45:00
        """
        round_trip(inp, exp)

    def test_explicit_t(self) -> None:  # to upper
        inp = """
        - 2011-10-02t16:45:00
        """
        exp = """
        - 2011-10-02T16:45:00
        """
        round_trip(inp, exp)

    def test_no_T_multi_space(self) -> None:
        inp = """
        - 2011-10-02   16:45:00
        """
        exp = """
        - 2011-10-02 16:45:00
        """
        round_trip(inp, exp)

    def test_iso(self) -> None:
        round_trip(
            """
        - 2011-10-02T15:45:00+01:00
        """
        )

    def test_zero_tz(self) -> None:
        round_trip(
            """
        - 2011-10-02T15:45:00+0
        """
        )

    def test_issue_45(self) -> None:
        round_trip(
            """
        dt: 2016-08-19T22:45:47Z
        """
        )

    def test_issue_366(self) -> None:
        import io

        import ruyaml

        round_trip(
            """
        [2021-02-01 22:34:48.696868-03:00]
        """
        )
        yaml = ruyaml.YAML()
        dd = DateTime(
            2021, 2, 1, 22, 34, 48, 696868, TimeZone(TimeDelta(hours=-3), name='')
        )
        buf = io.StringIO()
        yaml.dump(dd, buf)
        assert buf.getvalue() == '2021-02-01 22:34:48.696868-03:00\n...\n'
        rd = yaml.load(buf.getvalue())
        assert rd == dd

    def test_deepcopy_datestring(self) -> None:
        # reported by Quuxplusone, http://stackoverflow.com/a/41577841/1307905
        x = dedent(
            """\
        foo: 2016-10-12T12:34:56
        """
        )
        data = copy.deepcopy(round_trip_load(x))
        assert round_trip_dump(data) == x

    def test_fraction_overflow(self) -> None:
        # reported (indirectly) by Luís Ferreira
        # https://sourceforge.net/p/ruyaml/tickets/414/
        inp = dedent(
            """\
        - 2022-01-02T12:34:59.9999994
        - 2022-01-02T12:34:59.9999995
        """
        )
        exp = dedent(
            """\
        - 2022-01-02T12:34:59.999999
        - 2022-01-02T12:35:00
        """
        )
        round_trip(inp, exp)

    def Xtest_tzinfo(self) -> None:
        import ruyaml

        yaml = ruyaml.YAML()
        dts = '2011-10-02T16:45:00.930619+01:00'
        d = yaml.load(dts)
        print('d', repr(d), d._yaml)
        yaml.dump(dict(x=d), sys.stdout)
        print('----')
        # dx = DateTime.fromisoformat(dts)
        # print('dx', dx, repr(dx))
        dd = DateTime(
            2011,
            10,
            2,
            16,
            45,
            00,
            930619,
            TimeZone(TimeDelta(hours=1, minutes=0), name='+01:00'),
        )  # NOQA
        yaml.dump([dd], sys.stdout)
        print('dd', dd, dd.tzinfo)
        raise AssertionError()



"""
testing of anchors and the aliases referring to them
"""

from typing import Any

import pytest  # type: ignore  # NOQA
from roundtrip import (  # type: ignore # NOQA
    YAML,
    dedent,
    round_trip,
    round_trip_dump,
    round_trip_load,
)


def load(s: str) -> Any:
    return round_trip_load(dedent(s))


def compare(d: Any, s: str) -> None:
    assert round_trip_dump(d) == dedent(s)


class TestMerge:
    def test_remove_key_before_merge(self) -> None:
        data = load(
            """
        a: &aa
          b: 1
          c: 2
        d:
          e: 3
          f: 4
          <<: *aa
          g: 5
          h: 6
        """
        )
        del data['d']['f']
        compare(
            data,
            """
        a: &aa
          b: 1
          c: 2
        d:
          e: 3
          <<: *aa
          g: 5
          h: 6
        """,
        )

    def test_remove_key_after_merge(self) -> None:
        data = load(
            """
        a: &aa
          b: 1
          c: 2
        d:
          e: 3
          f: 4
          <<: *aa
          g: 5
          h: 6
        """
        )
        del data['d']['g']
        compare(
            data,
            """
        a: &aa
          b: 1
          c: 2
        d:
          e: 3
          f: 4
          <<: *aa
          h: 6
        """,
        )



# coding: utf-8

import os
import sys

import pytest  # type: ignore  # NOQA

sys.path.insert(0, os.path.dirname(__file__) + '/lib')

import warnings  # NOQA
from typing import Any, List  # NOQA

args: List[Any] = []


def test_data() -> None:
    import test_appliance  # type: ignore  # NOQA

    warnings.simplefilter('ignore', PendingDeprecationWarning)
    collections = []
    import test_yaml  # type: ignore

    collections.append(test_yaml)
    test_appliance.run(collections, args)


# @pytest.mark.skipif(not ruyaml.__with_libyaml__,
#                     reason="no libyaml")


def test_data_ext() -> None:
    collections = []
    import test_appliance  # NOQA

    import ruyaml

    warnings.simplefilter('ignore', ruyaml.error.UnsafeLoaderWarning)
    warnings.simplefilter('ignore', PendingDeprecationWarning)
    if ruyaml.__with_libyaml__:
        import test_yaml_ext  # type: ignore

        collections.append(test_yaml_ext)
        test_appliance.run(collections, args)



# coding: utf-8

import sys
import os
import warnings  # NOQA
from pathlib import Path
from typing import Any, List, Optional, Tuple

import pytest  # type: ignore  # NOQA

base_path = Path('data')  # that is ruyaml.data


class YAMLData:
    yaml_tag = '!YAML'

    def __init__(self, s: Any) -> None:
        self._s = s

    # Conversion tables for input. E.g. "<TAB>" is replaced by "\t"
    # fmt: off
    special = {
        'SPC': ' ',
        'TAB': '\t',
        '---': '---',
        '...': '...',
        'NL': '\n',
    }
    # fmt: on

    @property
    def value(self) -> Any:
        if hasattr(self, '_p'):
            return self._p  # type: ignore
        assert ' \n' not in self._s
        assert '\t\n' not in self._s
        self._p = self._s
        for k, v in YAMLData.special.items():
            k = '<' + k + '>'
            self._p = self._p.replace(k, v)
        return self._p

    def test_rewrite(self, s: str) -> str:
        assert ' \n' not in s
        assert '\t\n' not in s
        for k, v in YAMLData.special.items():
            k = '<' + k + '>'
            s = s.replace(k, v)
        return s

    @classmethod
    def from_yaml(cls, constructor: Any, node: Any) -> 'YAMLData':
        from ruyaml.nodes import MappingNode

        if isinstance(node, MappingNode):
            return cls(constructor.construct_mapping(node))
        return cls(node.value)


class Python(YAMLData):
    yaml_tag = '!Python'


class Output(YAMLData):
    yaml_tag = '!Output'


class Assert(YAMLData):
    yaml_tag = '!Assert'

    @property
    def value(self) -> Any:
        from collections.abc import Mapping

        if hasattr(self, '_pa'):
            return self._pa  # type: ignore
        if isinstance(self._s, Mapping):
            self._s['lines'] = self.test_rewrite(self._s['lines'])  # type: ignore
        self._pa = self._s
        return self._pa


class Events(YAMLData):
    yaml_tag = '!Events'


class JSONData(YAMLData):
    yaml_tag = '!JSON'


class Dump(YAMLData):
    yaml_tag = '!Dump'


class Emit(YAMLData):
    yaml_tag = '!Emit'


def pytest_generate_tests(metafunc: Any) -> None:
    test_yaml = []
    paths = sorted(base_path.glob('**/*.yaml'))
    idlist = []
    for path in paths:
        # while developing tests put them in data/debug and run:
        #   auto -c "pytest _test/test_z_data.py" data/debug/*.yaml *.py _test/*.py
        if os.environ.get('RUAMELAUTOTEST') == '1':
            if path.parent.stem != 'debug':
                continue
        elif path.parent.stem == 'debug':
            # don't test debug entries for production
            continue
        stem = path.stem
        if stem.startswith('.#'):  # skip emacs temporary file
            continue
        idlist.append(stem)
        test_yaml.append([path])
    metafunc.parametrize(['yaml'], test_yaml, ids=idlist, scope='class')


class TestYAMLData:
    def yaml(
        self,
        yaml_version: Optional[Any] = None,
        typ: Any = 'rt',
        pure: Any = None,
    ) -> Any:
        from ruyaml import YAML

        y = YAML(typ=typ, pure=pure)
        y.preserve_quotes = True
        if yaml_version:
            y.version = yaml_version
        y.composer.warn_double_anchors = False
        return y

    def docs(self, path: Path) -> List[Any]:
        from ruyaml import YAML

        tyaml = YAML(typ='safe', pure=True)
        tyaml.register_class(YAMLData)
        tyaml.register_class(Python)
        tyaml.register_class(Output)
        tyaml.register_class(Assert)
        tyaml.register_class(Events)
        tyaml.register_class(JSONData)
        tyaml.register_class(Dump)
        tyaml.register_class(Emit)
        return list(tyaml.load_all(path))

    def yaml_load(
        self, value: Any, yaml_version: Optional[Any] = None
    ) -> Tuple[Any, Any]:
        yaml = self.yaml(yaml_version=yaml_version)
        data = yaml.load(value)
        return yaml, data

    def round_trip(
        self,
        input: Any,
        output: Optional[Any] = None,
        yaml_version: Optional[Any] = None,
    ) -> None:
        from ruyaml.compat import StringIO

        yaml, data = self.yaml_load(input.value, yaml_version=yaml_version)
        buf = StringIO()
        yaml.dump(data, buf)
        expected = input.value if output is None else output.value
        value = buf.getvalue()
        print('>>>> rt output\n', value.replace(' ', '\u2423'), sep='')  # 2423 open box
        assert value == expected

    def gen_events(
        self,
        input: Any,
        output: Any,
        yaml_version: Optional[Any] = None,
    ) -> None:
        from ruyaml.compat import StringIO

        buf = StringIO()
        yaml = self.yaml(yaml_version=yaml_version)
        indent = 0
        try:
            for event in yaml.parse(input.value):
                compact = event.compact_repr()
                assert compact[0] in '+=-'
                if compact[0] == '-':
                    indent -= 1
                print(f'{" " * indent}{compact}', file=buf)
                if compact[0] == '+':
                    indent += 1

        except Exception as e:  # NOQA
            print('=EXCEPTION', file=buf)  # exceptions not indented
            if '=EXCEPTION' not in output.value:
                raise
        print('>>>> buf\n', buf.getvalue(), sep='')
        assert buf.getvalue() == output.value

    def load_compare_json(
        self,
        input: Any,
        output: Any,
        yaml_version: Optional[Any] = None,
    ) -> None:
        import json

        from ruyaml.comments import CommentedMap, TaggedScalar
        from ruyaml.compat import StringIO

        def serialize_obj(obj: Any) -> Any:
            if isinstance(obj, CommentedMap):
                return {k: v for k, v in obj.items()}  # NOQA
            elif isinstance(obj, TaggedScalar):
                return str(obj.value)
            elif isinstance(obj, set):
                return {k: None for k in obj}
            return str(obj)

        buf = StringIO()
        yaml = self.yaml(typ='rt', yaml_version=yaml_version)
        for data in yaml.load_all(input.value):
            if isinstance(data, dict):
                data = {str(k): v for k, v in data.items()}
            json.dump(data, buf, sort_keys=True, indent=2, default=serialize_obj)
            buf.write('\n')
        print('>>>> buf\n', buf.getvalue(), sep='')
        # jsons = json.dumps(json.loads(output.value))  # normalize formatting of JSON
        assert buf.getvalue() == output.value

    def load_compare_emit(
        self,
        input: Any,
        output: Any,
        yaml_version: Optional[Any] = None,
    ) -> None:
        from ruyaml.compat import StringIO

        buf = StringIO()
        yaml = self.yaml(yaml_version=yaml_version)
        yaml.preserve_quotes = True
        data = input.value
        if data.startswith('---') or '\n--- ' in data or '\n---' in data:
            yaml.explicit_start = True
        data = list(yaml.load_all(data))
        yaml.dump_all(data, buf)
        print('>>>> buf\n', buf.getvalue(), sep='')
        assert buf.getvalue() == output.value

    def load_assert(
        self,
        input: Any,
        confirm: Any,
        yaml_version: Optional[Any] = None,
    ) -> None:
        from collections.abc import Mapping

        d = self.yaml_load(input.value, yaml_version=yaml_version)[1]  # NOQA
        print('confirm.value', confirm.value, type(confirm.value))
        if isinstance(confirm.value, Mapping):
            r = range(confirm.value['range'])
            lines = confirm.value['lines'].splitlines()
            for idx in r:  # NOQA
                for line in lines:
                    line = 'assert ' + line
                    print(line)
                    exec(line)
        else:
            for line in confirm.value.splitlines():
                line = 'assert ' + line
                print(line)
                exec(line)

    def run_python(
        self,
        python: Any,
        data: Any,
        tmpdir: Any,
        input: Optional[Any] = None,
    ) -> None:
        from roundtrip import save_and_run  # type: ignore

        if input is not None:
            (tmpdir / 'input.yaml').write_text(input.value, encoding='utf-8')
        assert save_and_run(python.value, base_dir=tmpdir, output=data.value) == 0

    def insert_comments(self, data: Any, actions: Any) -> None:
        """this is to automatically insert based on:
        path (a.1.b),
        position (before, after, between), and
        offset (absolute/relative)
        """
        raise NotImplementedError
        expected = []
        for line in data.value.splitlines(True):
            idx = line.index['?']
            if idx < 0:
                expected.append(line)
                continue
            assert line.lstrip()[0] == '#'  # it has to be comment line
        print(data)
        assert ''.join(expected) == data.value

    # this is executed by pytest the methods with names not starting with
    # test_ are helper methods
    def test_yaml_data(self, yaml: Any, tmpdir: Any) -> None:
        from collections.abc import Mapping

        idx = 0
        typs = []  # list of test to be performed
        yaml_version = None

        docs = self.docs(yaml)
        if isinstance(docs[0], Mapping):
            d = docs[0]
            if d.get('skip'):
                pytest.skip('explicit skip')
            if '1.3-mod' in d.get('tags', []):
                pytest.skip('YAML 1.3')
            typ = d.get('type')
            if isinstance(typ, str):
                typs.append(typ)
            elif isinstance(typ, list):
                typs.extend(typ[:])
            del typ
            yaml_version = d.get('yaml_version')
            if 'python' in d:
                if not check_python_version(d['python']):
                    pytest.skip('unsupported version')
            idx += 1
        data = output = confirm = python = events = json = dump = emit = None
        for doc in docs[idx:]:
            if isinstance(doc, Output):
                output = doc
            elif isinstance(doc, Events):
                events = doc
            elif isinstance(doc, JSONData):
                json = doc
            elif isinstance(doc, Dump):
                dump = doc  # NOQA
            elif isinstance(doc, Emit):
                emit = doc  # NOQA
            elif isinstance(doc, Assert):
                confirm = doc
            elif isinstance(doc, Python):
                python = doc
                if len(typs) == 0:
                    typs = ['python_run']
            elif isinstance(doc, YAMLData):
                data = doc
            else:
                print('no handler for type:', type(doc), repr(doc))
                raise AssertionError()
        if len(typs) == 0:
            if data is not None and output is not None:
                typs = ['rt']
            elif data is not None and confirm is not None:
                typs = ['load_assert']
            else:
                assert data is not None
                typs = ['rt']
        print('type:', typs)
        if data is not None:
            print('>>>> data:\n', data.value.replace(' ', '\u2423'), sep='', end='')
        if events is not None:
            print('>>>> events:\n', events.value, sep='')
        else:
            print(
                '>>>> output:\n', output.value if output is not None else output, sep=''
            )
        for typ in typs:
            if typ == 'rt':
                self.round_trip(data, output, yaml_version=yaml_version)
            elif typ == 'python_run':
                inp = None if output is None or data is None else data
                self.run_python(
                    python,
                    output if output is not None else data,
                    tmpdir,
                    input=inp,
                )
            elif typ == 'load_assert':
                self.load_assert(data, confirm, yaml_version=yaml_version)
            elif typ == 'comment':
                actions: List[Any] = []
                self.insert_comments(data, actions)
            elif typ == 'events':
                if events is None:
                    print('need to specify !Events for type:', typ)
                    sys.exit(1)
                self.gen_events(data, events, yaml_version=yaml_version)
            elif typ == 'json':
                if json is None:
                    print('need to specify !JSON for type:', typ)
                    sys.exit(1)
                self.load_compare_json(data, json, yaml_version=yaml_version)
            elif typ == 'dump':
                continue
            elif typ == 'emit':
                self.load_compare_emit(data, emit)
            else:
                f'\n>>>>>> run type unknown: "{typ}" <<<<<<\n'
                raise AssertionError()


def check_python_version(match: Any, current: Optional[Any] = None) -> bool:
    """
    version indication, return True if version matches.
    match should be something like 3.6+, or [2.7, 3.3] etc. Floats
    are converted to strings. Single values are made into lists.
    """
    if current is None:
        current = list(sys.version_info[:3])
    if not isinstance(match, list):
        match = [match]
    for m in match:
        minimal = False
        if isinstance(m, float):
            m = str(m)
        if m.endswith('+'):
            minimal = True
            m = m[:-1]
        # assert m[0].isdigit()
        # assert m[-1].isdigit()
        m = [int(x) for x in m.split('.')]
        current_len = current[: len(m)]
        # print(m, current, current_len)
        if minimal:
            if current_len >= m:
                return True
        else:
            if current_len == m:
                return True
    return False



# coding: utf-8

#  try:
#      import numpy
#  except:  # NOQA
#      numpy = None


#  def Xtest_numpy() -> None:
#      import ruyaml
#
#      if numpy is None:
#          return
#      data = numpy.arange(10)
#      print('data', type(data), data)
#
#      buf = io.BytesIO()
#      ruyaml.dump(data)  # needs updating to use buffer
#      yaml_str = buf.getvalue().decode('utf-8')
#      datb = ruyaml.load(yaml_str)
#      print('datb', type(datb), datb)
#
#      print('\nYAML', yaml_str)
#      assert data == datb



from __future__ import annotations

from dataclasses import InitVar, dataclass, fields  # NOQA
from io import BytesIO
from textwrap import dedent
from typing import ClassVar, Union


class TestDataClasses:
    def test_1(self) -> None:
        from ruyaml import YAML

        yaml = YAML()

        @yaml.register_class
        @dataclass
        class DC:
            abc: int
            klm: int
            xyz: int = 0

            def __post_init__(self) -> None:
                self.xyz = self.abc + self.klm

        dc = DC(abc=5, klm=42)
        assert dc.xyz == 47

        yaml_str = dedent(
            """\
        !DC
        abc: 13
        klm: 37
        """
        )
        dc2 = yaml.load(yaml_str)
        assert dc2.xyz == 50

    def test_yamltag(self) -> None:
        from ruyaml import YAML

        yaml = YAML()

        @yaml.register_class
        @dataclass
        class DC:
            yaml_tag: ClassVar = '!dc_example'
            abc: int
            klm: int

        dc = DC(abc=5, klm=42)
        buf = BytesIO()
        yaml.dump(dc, buf)
        assert (
            buf.getvalue()
            == dedent(
                """\
        !dc_example
        abc: 5
        klm: 42
        """
            ).encode('utf-8')
        )
        dc2 = yaml.load(buf.getvalue())
        assert len(fields(dc2)) == 2  # class var is not a field
        assert dc2.abc == dc.abc
        assert dc2.klm == dc.klm

    def test_initvar(self) -> None:
        from ruyaml import YAML

        yaml = YAML()

        @yaml.register_class
        @dataclass
        class DC:
            abc: int
            klm: int
            xyz: InitVar[Union[str, None]] = None

            def __post_init__(self, xyz: Union[str, None]) -> None:
                # assert xyz == self.xyz  # self.xyz is always None
                if xyz is not None:
                    self.klm += len(xyz)

        dc = DC(abc=5, klm=42, xyz='provided')
        # this actually doesn't raise an attribute error, I would have expected it not to work
        # at all, but it has the default value
        assert dc.xyz is None  # type: ignore
        buf = BytesIO()
        yaml.dump(dc, buf)
        assert (
            buf.getvalue()
            == dedent(
                """\
        !DC
        abc: 5
        klm: 50
        """
            ).encode('utf-8')
        )

        yaml_str = dedent(
            """\
        !DC
        abc: 18
        klm: 55
        xyz: some string
        """
        )
        dc2 = yaml.load(yaml_str)
        assert dc2.xyz is None
        assert dc2.klm == 55 + len('some string')

    def test_initvar_not_in_yaml(self) -> None:
        from ruyaml import YAML

        yaml = YAML()

        @yaml.register_class
        @dataclass
        class DC:
            abc: int
            klm: int
            xyz: InitVar[Union[str, None]] = 'hello'

            def __post_init__(self, xyz: Union[str, None]) -> None:
                # assert xyz == self.xyz  # self.xyz is always None
                if xyz is not None:
                    self.klm += len(xyz)

        dc = DC(abc=5, klm=42, xyz='provided')
        assert dc.abc == 5
        assert dc.xyz == 'hello'  # type: ignore
        buf = BytesIO()
        yaml.dump(dc, buf)
        assert (
            buf.getvalue()
            == dedent(
                """\
        !DC
        abc: 5
        klm: 50
        """
            ).encode('utf-8')
        )

        yaml_str = dedent(
            """\
        !DC
        abc: 18
        klm: 55
        """
        )
        dc2 = yaml.load(yaml_str)
        assert dc2.xyz == 'hello'
        assert dc2.klm == 55 + len('hello')

    def test_collection_field(self) -> None:
        # https://stackoverflow.com/a/77485786/1307905
        from dataclasses import dataclass

        import ruyaml

        @dataclass
        class Msg:
            id: int
            desc: str
            fields: list[Field]

            def __post_init__(self) -> None:
                idx: int = 0
                for field in self.fields:  # why is this empty??
                    field.index = idx
                    idx += field.size

        @dataclass
        class Field:
            id: int
            name: str
            units: str
            size: int
            index: int = -1

        yaml = ruyaml.YAML()
        yaml.register_class(Msg)
        yaml.register_class(Field)

        msg: Msg = yaml.load(
            """\
        !Msg
        id: 1
        desc: status
        fields:
        - !Field
            id: 1
            name: Temp
            units: degC
            size: 2
        """
        )

        assert msg.fields[0].index != -1
